`include "instructions.v"

module Beepo #(
    parameter FREQ      = 27_000_000, 
    parameter UART_BAUD = 115200,
    parameter WORD_SIZE = 16
) (
    input i_clk,
    input i_button1,
    input i_resume,
    output [6:0] o_segments_drive,
    output [3:0] o_displays_neg,
    output o_breakpoint,
    output o_uart_tx
);
    // State values
    localparam  IDLE    = 0;    // Start fetching instruction
    localparam  FETCHI  = 1;    // Instruction is fetched, start fetching first argument
    localparam  FETCHA  = 2;    // Argument byte is fetched
    localparam  EXEC    = 3;    // Start running
    localparam  MEMR    = 4;    // Transferring bytes between memory and registers
    localparam  DONE    = 7;    // Done executing

    // Argument types
    localparam [3:0] ARG_R   = 0;   // Register
    localparam [3:0] ARG_O   = 1;   // 32 bit relative PC offset
    localparam [3:0] ARG_P   = 2;   // 16 bit relative PC offset
    localparam [3:0] ARG_B   = 3;   // Byte, 8 bit
    localparam [3:0] ARG_H   = 4;   // Half-word, 16 bit
    localparam [3:0] ARG_W   = 5;   // Word, 32 bit
    localparam [3:0] ARG_D   = 6;   // Double-word, 64 bit
    localparam [3:0] ARG_A   = 7;   // Absolute address immediate, 64 bit
    localparam [3:0] ARG_N   = 8;   // No argument

    localparam [0:31] ARG_SIZES = {4'd1, 4'd4, 4'd2, 4'd1, 4'd2, 4'd4, 4'd8, 4'd8};

    localparam  PC_START    = 0;
    localparam  NUM_REGS    = 4;

    reg [2:0]   r_state     = IDLE;
    reg [WORD_SIZE-1:0]  r_tick      = 0;

    // Registers
    reg [WORD_SIZE-1:0] r_pc        = PC_START; // program counter
    reg [WORD_SIZE-1:0] r_registers [0:NUM_REGS];   // up to 255 modifiable registers
    reg [7:0]           r_instr;                // the current instruction
    reg [7:0]           r_arg_regs  [0:3];      // register arguments
    reg [WORD_SIZE-1:0] r_arg_imm   = 0;        // immediate argument
    reg [WORD_SIZE-1:0] r_arg_addr  = 0;        // relative/absolute address argument

    reg [1:0]   r_arg_index         = 3;    // the instruction index currently being fetched
    reg [3:0]   r_arg_types         [0:3];  // the types of each argument to be fetched
    reg [15:0]  r_arg_types_packed  = 0;    // to be unpacked into r_arg_types
    reg [2:0]   r_arg_bytes         = 0;    // the number of bytes left to fetch for the current argument
    reg [3:0]   r_arg_current_type  = 8;    // the type of the current argument
    reg [5:0]   r_arg_bit           = 0;    // the current lower bit index being fetched for the current argument

    reg         r_mem_wre   = 0;
    reg         r_mem_busy  = 0;
    reg         r_mem_trans = 0;
    reg [7:0]   r_mem_in    = 0;
    reg [7:0]   r_mem_index = 0;    // the index of the byte in transfer
    reg [7:0]   r_mem_reg   = 0;    // the register currently used in transfer
    wire [WORD_SIZE-1:0] w_mem_addr  = r_mem_trans ? r_arg_addr : r_pc;
    wire [7:0]  w_mem_fetch;

    reg         r_breakpoint = 0;
    assign      o_breakpoint = r_breakpoint;

    reg                 r_uart_rst_n    = 0;
    reg [7:0]           r_uart_data     = 0;
    reg                 r_uart_busy     = 0;
    reg [7:0]           r_uart_left     = 0;
    reg [WORD_SIZE-1:0] r_uart_addr     = 0;
    wire w_uart_ready;
    
    genvar i;

    generate
        for (i = 0; i <= NUM_REGS; i = i + 1) begin
            initial r_registers[i] <= 0;
        end
    endgenerate

    always @(posedge i_clk) r_tick <= r_tick + 1;

    always @(posedge i_clk) begin
        if (r_breakpoint == 1) begin
            r_breakpoint = ~i_resume;
        end else if (r_mem_busy == 1) begin
            r_mem_busy = 0;
        end else case (r_state)
            IDLE: begin
                r_state <= FETCHI;
                r_mem_wre <= 0;
                r_mem_busy <= 1;
            end
            FETCHI: begin
                r_instr <= w_mem_fetch;
                r_arg_index <= 0; 
                r_arg_bit <= 0;
                r_mem_trans <= 0;


                case (w_mem_fetch)
                    `TX:        r_arg_types_packed = `TX_ARGS;
                    `NOP:       r_arg_types_packed = `NOP_ARGS;
                    `ADD8:      r_arg_types_packed = `ADD8_ARGS;
                    `ADD16:     r_arg_types_packed = `ADD16_ARGS;
                    `ADD32:     r_arg_types_packed = `ADD32_ARGS;
                    `ADD64:     r_arg_types_packed = `ADD64_ARGS;
                    `ADDI8:     r_arg_types_packed = `ADDI8_ARGS;
                    `ADDI16:    r_arg_types_packed = `ADDI16_ARGS;
                    `ADDI32:    r_arg_types_packed = `ADDI32_ARGS;
                    `ADDI64:    r_arg_types_packed = `ADDI64_ARGS;
                    `LI8:       r_arg_types_packed = `LI8_ARGS;
                    `LI16:      r_arg_types_packed = `LI16_ARGS;
                    `LI32:      r_arg_types_packed = `LI32_ARGS;
                    `LI64:      r_arg_types_packed = `LI64_ARGS;
                    `LD:        r_arg_types_packed = `LD_ARGS;
                    `ST:        r_arg_types_packed = `ST_ARGS;
                    `EBP:       r_arg_types_packed = `EBP_ARGS;
                    default:    r_arg_types_packed = {ARG_N, ARG_N, ARG_N, ARG_N};
                endcase

                r_pc = r_pc + 1;
                r_mem_busy = 1;

                if (r_arg_types_packed[15:12] != ARG_N) begin
                    r_state <= FETCHA;

                    r_arg_bytes <= ARG_SIZES[r_arg_types_packed[15:12]*4+:4];
                    r_arg_current_type <= r_arg_types_packed[15:12];

                    r_arg_types[0] <= r_arg_types_packed[15:12];
                    r_arg_types[1] <= r_arg_types_packed[11:8];                    
                    r_arg_types[2] <= r_arg_types_packed[7:4];
                    r_arg_types[3] <= r_arg_types_packed[3:0];

                    r_arg_regs[0] <= 0;
                    r_arg_regs[1] <= 0;
                    r_arg_regs[2] <= 0;
                    r_arg_regs[3] <= 0;
                    r_arg_imm <= 0;
                    r_arg_addr <= 0;
                end else r_state <= EXEC;
            end
            FETCHA: begin
                if (r_arg_current_type == ARG_N) r_state <= IDLE;
                else begin
                    case (r_arg_current_type)
                        ARG_R: r_arg_regs[r_arg_index]  <= w_mem_fetch;
                        ARG_O: r_arg_addr[r_arg_bit+:8] <= w_mem_fetch;
                        ARG_P: r_arg_addr[r_arg_bit+:8] <= w_mem_fetch;
                        ARG_B: r_arg_imm[r_arg_bit+:8]  <= w_mem_fetch;
                        ARG_H: r_arg_imm[r_arg_bit+:8]  <= w_mem_fetch;
                        ARG_W: r_arg_imm[r_arg_bit+:8]  <= w_mem_fetch;
                        ARG_D: r_arg_imm[r_arg_bit+:8]  <= w_mem_fetch;
                        ARG_A: r_arg_addr[r_arg_bit+:8] <= w_mem_fetch;
                    endcase

                    r_pc <= r_pc + 1;
                    r_mem_busy <= 1;
                    
                    r_arg_bytes = r_arg_bytes - 1;
                    r_arg_bit <= r_arg_bit + 8;
                    
                    if (r_arg_bytes == 0) begin
                        r_arg_index = r_arg_index + 1;
                        r_arg_current_type = r_arg_types[r_arg_index];

                        // Execute when there is no next argument or r_arg_index has overflowed
                        if (r_arg_current_type == ARG_N || r_arg_index == 0) r_state <= EXEC;
                        else begin
                            r_arg_bit   <= 0;
                            r_arg_bytes <= ARG_SIZES[r_arg_current_type*4+:4];
                        end
                    end
                end
            end
            EXEC: begin
                r_state <= FETCHI;

                if (WORD_SIZE >= 8)
                    case (r_instr)
                        `TX:        r_state <= DONE;
                        `NOP:       ;
                        `ADD8:      set_reg_byte    (r_arg_regs[0], r_registers[r_arg_regs[1]][0+:8]    + r_registers   [r_arg_regs[2]][0+:8]);
                        `ADDI8:     set_reg_byte    (r_arg_regs[0], r_registers[r_arg_regs[1]][0+:8]    + r_arg_imm     [0+:8]);
                        `LI8:       set_reg_byte    (r_arg_regs[0], r_arg_imm);
                        `LD: begin
                            if (r_arg_imm > 0) begin
                                r_arg_addr  <= r_arg_addr + r_registers[r_arg_regs[1]];
                                r_mem_index <= 0;
                                r_mem_reg   <= r_arg_regs[0];
                                r_mem_busy  <= 1;
                                r_state     <= MEMR;
                                r_mem_trans <= 1;
                            end
                        end
                        `ST: begin
                            if (r_arg_imm > 0) begin
                                r_arg_addr  <= r_arg_addr + r_registers[r_arg_regs[1]];
                                r_mem_index <= 1;
                                r_mem_reg   <= r_arg_regs[0];
                                r_mem_wre   <= 1;
                                r_mem_in    <= r_registers[r_arg_regs[0]][0+:8];
                                r_mem_busy  <= 1;
                                r_state     <= MEMR;
                                r_mem_trans <= 1;
                            end
                        end
                        `EBP: r_breakpoint = 1;
                    endcase

                if (WORD_SIZE >= 16)
                    case (r_instr)
                        `ADD16:     set_reg_hword   (r_arg_regs[0], r_registers[r_arg_regs[1]][0+:16]   + r_registers   [r_arg_regs[2]][0+:16]);
                        `ADDI16:    set_reg_hword   (r_arg_regs[0], r_registers[r_arg_regs[1]][0+:16]   + r_arg_imm     [0+:16]);
                        `LI16:      set_reg_hword   (r_arg_regs[0], r_arg_imm);
                    endcase

                if (WORD_SIZE >= 32)
                    case (r_instr)
                        `ADD32:     set_reg_word    (r_arg_regs[0], r_registers[r_arg_regs[1]][0+:32]   + r_registers   [r_arg_regs[2]][0+:32]);
                        `ADDI32:    set_reg_word    (r_arg_regs[0], r_registers[r_arg_regs[1]][0+:32]   + r_arg_imm     [0+:32]);
                        `LI32:      set_reg_word    (r_arg_regs[0], r_arg_imm);
                    endcase

                if (WORD_SIZE >= 64)
                    case (r_instr)
                        `ADD64:     set_reg_dword   (r_arg_regs[0], r_registers[r_arg_regs[1]][0+:64]   + r_registers   [r_arg_regs[2]][0+:64]);
                        `ADDI64:    set_reg_dword   (r_arg_regs[0], r_registers[r_arg_regs[1]][0+:64]   + r_arg_imm     [0+:64]);
                        `LI64:      set_reg_dword   (r_arg_regs[0], r_arg_imm);
                    endcase
            end
            MEMR: begin
                case (r_instr)
                    `LD:    set_reg_part(r_mem_reg, w_mem_fetch, r_mem_index*8);
                    `ST:    r_mem_in <= r_registers[r_mem_reg][r_mem_index*8+:8];
                endcase                

                r_mem_busy <= 1;

                if (r_arg_imm == 1) begin
                    // reached the end of the transfer
                    r_mem_wre <= 0;
                    r_mem_trans <= 0;
                    r_state <= FETCHI;
                end else begin
                    r_mem_index = r_mem_index + 1;

                    if (r_mem_index == WORD_SIZE / 8) begin
                        // reached the end of this register
                        r_mem_reg <= r_mem_reg + 1;
                        r_mem_index <= 0;
                    end

                    r_arg_addr  <= r_arg_addr + 1;
                    r_arg_imm   <= r_arg_imm - 1;
                end
            end
        endcase

        r_registers[0] <= 0;
    end

    task automatic set_reg_byte(
        input [7:0] being_set,
        input [7:0] setting_to
    );
        r_registers[being_set][0+:8] = setting_to;
    endtask

    task automatic set_reg_hword(
        input [7:0] being_set,
        input [15:0] setting_to
    );
        r_registers[being_set][0+:16] = setting_to;
    endtask

    task automatic set_reg_word(
        input [7:0] being_set,
        input [31:0] setting_to
    );
        r_registers[being_set][0+:32] = setting_to;
    endtask

    task automatic set_reg_dword(
        input [7:0] being_set,
        input [WORD_SIZE-1:0] setting_to
    );
        r_registers[being_set][0+:64] = setting_to;
    endtask

    task automatic set_reg_part(
        input [7:0] being_set,
        input [7:0] setting_to,
        input [5:0] start_bit
    );
        if (start_bit <= 56) r_registers[being_set][start_bit+:8] = setting_to;
    endtask

   Multi7 display (
       .i_clk(i_clk),
       .i_hex({r_registers[1][15:0]}),
       .o_segments_drive(o_segments_drive),
       .o_displays_neg(o_displays_neg)
   );

    // uart_tx #(
    //     .CLK_FRE(FREQ/1_000_000),
    //     .BAUD_RATE(115200)
    // ) serial (
    //     .clk(i_clk),
    //     .rst_n(r_uart_rst_n),
    //     .tx_data(r_uart_data),
    //     .tx_data_ready(w_uart_ready),
    //     .tx_pin(o_uart_tx)
    // );

    spMem mem (
        .clk(i_clk),
        .ad(w_mem_addr),
        .din(r_mem_in),
        .dout(w_mem_fetch),
        .oce(0),
        .ce(1),
        .reset(0),
        .wre(r_mem_wre)
    );
endmodule