// To read:
// 1. Set i_addr to start address
// 2. Set i_flags.0 to 0
// 3. Set i_size to number of bytes to read
// 4. Pulse i_start high
// 5. When o_ready goes high, the read data will be in o_out

// To write:
// 1. Set i_addr to start address
// 2. Set i_in to the data to write
// 3. Set i_flags.0 to 1
// 4. Pulse i_start high
// 5. When o_ready goes high, the transfer is complete
module Bus#(
    parameter ADDR_WIDTH = 16,
    parameter DATA_WIDTH = 256
) (
    input i_clk,
    input [ADDR_WIDTH-1:0] i_addr,
    input [DATA_WIDTH-1:0] i_in,
    input [0:0] i_flags,  // flags.0: read(0)/write(1)
    input [5:0] i_size,
    input i_start,  // pulsed high to start transfer
    output o_ready,
    output [DATA_WIDTH-1:0] o_out
);
    localparam F_READ   = 'b0;
    localparam F_WRITE  = 'b1;

    localparam S_IDLE   = 0;
    localparam S_BUSY   = 1;

    reg r_status = S_IDLE;
    reg r_enable = 0;
    reg [5:0] r_tx_size     = 0;
    reg [5:0] r_byte_index  = 0;
    reg [7:0] r_in = 0;
    reg [ADDR_WIDTH-1:0] r_mem_addr;

    assign o_ready = r_byte_index == 0 || r_byte_index > r_tx_size;

    always @(posedge i_clk or posedge i_start) begin
        if (i_start && !r_enable) begin
            r_mem_addr  <= i_addr;
            r_enable    <= 1;
            r_status    <= S_BUSY;
            r_tx_size   <= i_size;
            r_in        <= i_in[0+:8];
            r_byte_index <= 1; // 0 is transferring now
        end else if (r_status == S_BUSY) r_status <= S_IDLE;
        else if (r_enable && r_byte_index > r_tx_size) r_enable <= 0;
        else if (r_enable) begin
            // increment address, input next byte
            r_mem_addr  <= r_mem_addr + 1;
            r_status    <= S_BUSY;
            r_in        <= i_in[r_byte_index*8+:8];
            r_byte_index <= r_byte_index + 1;
        end
    end

    spMem memory (
        .dout(o_out), //output [7:0] dout
        .clk(i_clk),    //input clk
        .oce(1'b0),     //input oce (unused)
        .ce(r_enable),      //input ce
        .reset(1'b0),   //input reset
        .wre(i_flags[0]),   //input wre (write enable)
        .ad(r_mem_addr),    //input [15:0] ad
        .din(r_in)      //input [7:0] din
    );
endmodule